==> podxtpro-0.3.tbz: podxtpro-0.3/playback.c <==
/*
 * Line6 PODxt Pro USB driver - 0.3
 *
 * Copyright (C) 2004, 2005 Markus Grabner (grabner@icg.tu-graz.ac.at)
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License as
 *	published by the Free Software Foundation, version 2.
 *
 */

#include <linux/config.h>

#ifdef CONFIG_USB_DEBUG
#define DEBUG 1
#endif

#include <asm/uaccess.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/usb.h>
#include <linux/time.h>

#include <sound/driver.h>
#include <sound/core.h>
#include <sound/control.h>
#include <sound/initval.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>

#include "audio.h"
#include "config.h"
#include "driver.h"


#define chip_t snd_podxtpro_audio_t


/* hardware definition (playback) */
static snd_pcm_hardware_t snd_podxtpro_playback_hw = {
	.info = (SNDRV_PCM_INFO_MMAP |
					 SNDRV_PCM_INFO_INTERLEAVED |
					 SNDRV_PCM_INFO_BLOCK_TRANSFER |
					 SNDRV_PCM_INFO_MMAP_VALID),
	.formats =          SNDRV_PCM_FMTBIT_S24_3LE,
	.rates =            SNDRV_PCM_RATE_KNOT,
	.rate_min =         39063,
	.rate_max =         39063,
	.channels_min =     2,
	.channels_max =     2,
	.buffer_bytes_max = 60000,
#if ARDOUR_COMPAT
	.period_bytes_min = 256,
	.period_bytes_max = 8192,
#else
	.period_bytes_min = PODXTPRO_FRAMES_PER_PERIOD * PODXTPRO_BYTES_PER_FRAME,
	.period_bytes_max = PODXTPRO_FRAMES_PER_PERIOD * PODXTPRO_BYTES_PER_FRAME,
#endif
	.periods_min =      1,
	.periods_max =      1024,
};


/*
	Software stereo volume control.
*/
static void change_volume(struct urb *urb_out, int volume[])
{
	int chn = 0;
	unsigned char *p, *buf_end;

	if(volume[0] == 256 && volume[1] == 256)
		return;  /* maximum volume - no change */

	p = (unsigned char *)urb_out->transfer_buffer;
	buf_end = p + urb_out->transfer_buffer_length;

	for(; p < buf_end; p += 3) {
		int val = p[0] + (p[1] << 8) + ((signed char)p[2] << 16);
		val = (val * volume[chn & 1]) >> 8;
		p[0] = val;
		p[1] = val >> 8;
		p[2] = val >> 16;
		++chn;
	}
}

/*
	Find a free URB, prepare audio data, and submit URB.
*/
static int submit_audio_out_urb(snd_pcm_substream_t *substream)
{
	int index;
	unsigned long flags;
	snd_podxtpro_audio_t *chip = snd_pcm_substream_chip(substream);
	snd_pcm_runtime_t *runtime = substream->runtime;

	spin_lock_irqsave(&chip->lock_audio_out, flags);
	index = find_first_zero_bit(&chip->active_urb_out, PODXTPRO_ISO_BUFFERS);

	if(index < 0 || index >= PODXTPRO_ISO_BUFFERS) {
		spin_unlock_irqrestore(&chip->lock_audio_out, flags);
		dev_err(s2m(substream), "No free URB found\n");
		return -EINVAL;
	}

	int i, urb_size, urb_frames;
	struct urb *urb_out = chip->urb_audio_out[index];
	urb_size = 0;

	for(i = 0; i < PODXTPRO_ISO_PACKETS; ++i) {
		int fs = (index == 0 && i == 0) ? PODXTPRO_ISO_PACKET_SIZE_LONG : PODXTPRO_ISO_PACKET_SIZE_NORMAL;
		struct usb_iso_packet_descriptor *fout = &urb_out->iso_frame_desc[i];
		fout->offset = urb_size;
		fout->length = fs;
		urb_size += fs;
	}

	urb_frames = urb_size / PODXTPRO_BYTES_PER_FRAME;

	if(chip->pos_out + urb_frames > runtime->buffer_size) {
		/*
			The transferred area goes over buffer boundary,
			copy the data to the temp buffer.
		*/
		int len;
		len = runtime->buffer_size - chip->pos_out;
		urb_out->transfer_buffer = chip->wrap_out;
		memcpy(chip->wrap_out, runtime->dma_area + chip->pos_out * PODXTPRO_BYTES_PER_FRAME, len * PODXTPRO_BYTES_PER_FRAME);
		memcpy(chip->wrap_out + len * PODXTPRO_BYTES_PER_FRAME, runtime->dma_area, (urb_frames - len) * PODXTPRO_BYTES_PER_FRAME);
	}
	else {
		/* set the buffer pointer */
		urb_out->transfer_buffer = runtime->dma_area + chip->pos_out * PODXTPRO_BYTES_PER_FRAME;
	}

	if((chip->pos_out += urb_frames) >= runtime->buffer_size)
		chip->pos_out -= runtime->buffer_size;

	urb_out->transfer_buffer_length = urb_size;
	urb_out->context = substream;
	change_volume(urb_out, chip->volume);

	if(usb_submit_urb(urb_out, GFP_ATOMIC) == 0)
		set_bit(index, &chip->active_urb_out);
	else
		dev_err(s2m(substream), "URB out #%d submission failed\n", index);
	
	spin_unlock_irqrestore(&chip->lock_audio_out, flags);
	return 0;
}

/*
	Submit all currently available playback URBs.
*/
static int submit_audio_out_all_urbs(snd_pcm_substream_t *substream)
{
	int ret, i;

	for(i = 0; i < PODXTPRO_ISO_BUFFERS; ++i)
		if((ret = submit_audio_out_urb(substream)) < 0)
			return ret;

	return 0;
}

/*
	Unlink all currently active playback URBs.
*/
static void unlink_audio_out_urbs(snd_podxtpro_audio_t *chip)
{
	unsigned int i;

	for(i = PODXTPRO_ISO_BUFFERS; i--;) {
		if(test_bit(i, &chip->active_urb_out)) {
			if(!test_and_set_bit(i, &chip->unlink_urb_out)) {
				struct urb *u = chip->urb_audio_out[i];
				u->transfer_flags |= URB_ASYNC_UNLINK;
				usb_unlink_urb(u);
			}
		}
	}
}

/*
	Wait until unlinking of all currently active playback URBs has been	finished.
*/
static void wait_clear_audio_out_urbs(snd_podxtpro_audio_t *chip)
{
	int timeout = HZ;
	unsigned int i;
	int alive;

	do {
		alive = 0;
		for (i = PODXTPRO_ISO_BUFFERS; i--;) {
			if (test_bit(i, &chip->active_urb_out))
				alive++;
		}
		if (! alive)
			break;
		set_current_state(TASK_UNINTERRUPTIBLE);
		schedule_timeout(1);
	} while (--timeout > 0);
	if (alive)
		snd_printk(KERN_ERR "timeout: still %d active urbs..\n", alive);

	chip->active_urb_out = 0;
	chip->unlink_urb_out = 0;
}

/*
	Unlink all currently active playback URBs, and wait for finishing.
*/
void unlink_wait_clear_audio_out_urbs(snd_podxtpro_audio_t *chip)
{
	unlink_audio_out_urbs(chip);
	wait_clear_audio_out_urbs(chip);
}

/*
	Callback for completed playback URB.
*/
static void audio_out_callback(struct urb *urb, struct pt_regs *regs)
{
	int i, index, length = 0, shutdown = 0;
	unsigned long flags;

	snd_pcm_substream_t *substream = (snd_pcm_substream_t *)urb->context;
	snd_podxtpro_audio_t *chip = snd_pcm_substream_chip(substream);
	snd_pcm_runtime_t *runtime = substream->runtime;

	/* find index of URB */
	for(index = PODXTPRO_ISO_BUFFERS; index--;)
		if(urb == chip->urb_audio_out[index])
			break;

	for(i = PODXTPRO_ISO_PACKETS; i--;)
		length += urb->iso_frame_desc[i].length;

	spin_lock_irqsave(&chip->lock_audio_out, flags);
	chip->pos_out_done += length / PODXTPRO_BYTES_PER_FRAME;

	if(chip->pos_out_done >= runtime->buffer_size)
		chip->pos_out_done -= runtime->buffer_size;
	
	clear_bit(index, &chip->active_urb_out);

	for(i = PODXTPRO_ISO_PACKETS; i--;)
		if(urb->iso_frame_desc[i].status == -ESHUTDOWN) {
			shutdown = 1;
			break;
		}

	if(test_bit(index, &chip->unlink_urb_out))  /* better use a chip->flag! */
		shutdown = 1;

	spin_unlock_irqrestore(&chip->lock_audio_out, flags);

	if(!shutdown) {
		submit_audio_out_urb(substream);

		if((chip->bytes_out += length) >= chip->period_out) {
			chip->bytes_out -= chip->period_out;
			snd_pcm_period_elapsed(substream);
		}
	}

	// return 0;
}

/* open playback callback */
static int snd_podxtpro_playback_open(snd_pcm_substream_t *substream)
{
	snd_pcm_runtime_t *runtime = substream->runtime;
	runtime->hw = snd_podxtpro_playback_hw;
	return 0;
}

/* close playback callback */
static int snd_podxtpro_playback_close(snd_pcm_substream_t *substream)
{
	return 0;
}

/* hw_params playback callback */
static int snd_podxtpro_playback_hw_params(snd_pcm_substream_t *substream, snd_pcm_hw_params_t *hw_params)
{
	int ret;

	snd_podxtpro_audio_t *chip = snd_pcm_substream_chip(substream);

	if((ret = snd_pcm_lib_malloc_pages(substream, params_buffer_bytes(hw_params))) < 0)
		return ret;

	chip->period_out = params_period_bytes(hw_params);
	// dev_info(s2m(substream), "podxtpro period out size = %d\n", chip->period_out);
	chip->wrap_out = kmalloc(PODXTPRO_ISO_PACKET_SIZE_NORMAL + PODXTPRO_ISO_PACKET_SIZE_LONG, GFP_KERNEL);

	if(!chip->wrap_out) {
		dev_err(s2m(substream), "cannot malloc wrap_out\n");
		return -ENOMEM;
	}

	return 0;
}

/* hw_free playback callback */
static int snd_podxtpro_playback_hw_free(snd_pcm_substream_t *substream)
{
	snd_podxtpro_audio_t *chip = snd_pcm_substream_chip(substream);
	unlink_wait_clear_audio_out_urbs(chip);

	if(chip->wrap_out) {
		kfree(chip->wrap_out);
		chip->wrap_out = 0;
	}

	return snd_pcm_lib_free_pages(substream);
}

/* prepare playback callback */
static int snd_podxtpro_playback_prepare(snd_pcm_substream_t *substream)
{
	snd_podxtpro_audio_t *chip = snd_pcm_substream_chip(substream);
	chip->pos_out = 0;
	chip->pos_out_done = 0;
	chip->bytes_out = 0;
	unlink_wait_clear_audio_out_urbs(chip);
	return 0;
}

/* trigger playback callback */
static int snd_podxtpro_playback_trigger(snd_pcm_substream_t *substream, int cmd)
{
	snd_podxtpro_audio_t *chip = snd_pcm_substream_chip(substream);

	switch (cmd) {
	case SNDRV_PCM_TRIGGER_START:
		submit_audio_out_all_urbs(substream);
		break;

	case SNDRV_PCM_TRIGGER_STOP:
		unlink_audio_out_urbs(chip);
		break;

	default:
		return -EINVAL;
	}

	return 0;
}

/* playback pointer callback */
static snd_pcm_uframes_t
snd_podxtpro_playback_pointer(snd_pcm_substream_t *substream)
{
	snd_podxtpro_audio_t *chip = snd_pcm_substream_chip(substream);
	return chip->pos_out_done;
}

/* playback operators */
snd_pcm_ops_t snd_podxtpro_playback_ops = {
	.open =        snd_podxtpro_playback_open,
	.close =       snd_podxtpro_playback_close,
	.ioctl =       snd_pcm_lib_ioctl,
	.hw_params =   snd_podxtpro_playback_hw_params,
	.hw_free =     snd_podxtpro_playback_hw_free,
	.prepare =     snd_podxtpro_playback_prepare,
	.trigger =     snd_podxtpro_playback_trigger,
	.pointer =     snd_podxtpro_playback_pointer,
};

int create_audio_out_urbs(snd_podxtpro_audio_t *chip)
{
	int i;
	usb_podxtpro_t *podxtpro = chip->podxtpro;

/* create audio URBs and fill in constant values: */
	for(i = 0; i < PODXTPRO_ISO_BUFFERS; ++i)
		{
			struct urb *urb;

			/* URB for audio out: */
			urb = chip->urb_audio_out[i] = usb_alloc_urb(PODXTPRO_ISO_PACKETS, GFP_KERNEL);

			if(urb == NULL) {
				dev_err(podxtpro->msgdev, "Out of memory\n");
				return -ENOMEM;
			}

			urb->dev = podxtpro->usbdev;
			urb->pipe = usb_sndisocpipe(podxtpro->usbdev, PODXTPRO_EP_WRITE_AUDIO & USB_ENDPOINT_NUMBER_MASK);
			urb->transfer_flags = URB_ISO_ASAP;
			urb->start_frame = -1;
			urb->number_of_packets = PODXTPRO_ISO_PACKETS;
			urb->interval = PODXTPRO_ISO_INTERVAL;
			urb->error_count = 0;
			urb->complete = audio_out_callback;
		}

	return 0;
}

==> podxtpro-0.4.tbz: podxtpro-0.4/playback.c <==
/*
 * Line6 PODxt Pro USB driver - 0.4
 *
 * Copyright (C) 2004, 2005 Markus Grabner (grabner@icg.tu-graz.ac.at)
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License as
 *	published by the Free Software Foundation, version 2.
 *
 */

#include <linux/config.h>

#ifdef CONFIG_USB_DEBUG
#define DEBUG 1
#endif

#include <asm/uaccess.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/usb.h>
#include <linux/time.h>

#include <sound/driver.h>
#include <sound/core.h>
#include <sound/control.h>
#include <sound/initval.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>

#include "audio.h"
#include "config.h"
#include "driver.h"


/* hardware definition (playback) */
static snd_pcm_hardware_t snd_podxtpro_playback_hw = {
	.info = (SNDRV_PCM_INFO_MMAP |
					 SNDRV_PCM_INFO_INTERLEAVED |
					 SNDRV_PCM_INFO_BLOCK_TRANSFER |
					 SNDRV_PCM_INFO_MMAP_VALID |
					 SNDRV_PCM_INFO_SYNC_START),
	.formats =          SNDRV_PCM_FMTBIT_S24_3LE,
	.rates =            SNDRV_PCM_RATE_KNOT,
	.rate_min =         39063,
	.rate_max =         39063,
	.channels_min =     2,
	.channels_max =     2,
	.buffer_bytes_max = 60000,
	.period_bytes_min = PODXTPRO_ISO_PACKET_SIZE_LONG * PODXTPRO_BYTES_PER_FRAME,  /* at least the longest URB must fit into one period */
	.period_bytes_max = 8192,
	.periods_min =      1,
	.periods_max =      1024,
};


/*
	Software stereo volume control.
*/
static void change_volume(struct urb *urb_out, int volume[])
{
	int chn = 0;
	unsigned char *p, *buf_end;

	if(volume[0] == 256 && volume[1] == 256)
		return;  /* maximum volume - no change */

	p = (unsigned char *)urb_out->transfer_buffer;
	buf_end = p + urb_out->transfer_buffer_length;

	for(; p < buf_end; p += 3) {
		int val = p[0] + (p[1] << 8) + ((signed char)p[2] << 16);
		val = (val * volume[chn & 1]) >> 8;
		p[0] = val;
		p[1] = val >> 8;
		p[2] = val >> 16;
		++chn;
	}
}

/*
	Find a free URB, prepare audio data, and submit URB.
*/
static int submit_audio_out_urb(snd_pcm_substream_t *substream)
{
	int index;
	unsigned long flags;
	snd_podxtpro_audio_t *chip = snd_pcm_substream_chip(substream);
	snd_pcm_runtime_t *runtime = substream->runtime;

	spin_lock_irqsave(&chip->lock_audio_out, flags);
	index = find_first_zero_bit(&chip->active_urb_out, PODXTPRO_ISO_BUFFERS);

	if(index < 0 || index >= PODXTPRO_ISO_BUFFERS) {
		spin_unlock_irqrestore(&chip->lock_audio_out, flags);
		dev_err(s2m(substream), "No free URB found\n");
		return -EINVAL;
	}

	int i, urb_size, urb_frames;
	struct urb *urb_out = chip->urb_audio_out[index];
	urb_size = 0;

	for(i = 0; i < PODXTPRO_ISO_PACKETS; ++i) {
		int fs = (index == 0 && i == 0) ? PODXTPRO_ISO_PACKET_SIZE_LONG : PODXTPRO_ISO_PACKET_SIZE_NORMAL;
		struct usb_iso_packet_descriptor *fout = &urb_out->iso_frame_desc[i];
		fout->offset = urb_size;
		fout->length = fs;
		urb_size += fs;
	}

	urb_frames = urb_size / PODXTPRO_BYTES_PER_FRAME;

	if(chip->pos_out + urb_frames > runtime->buffer_size) {
		/*
			The transferred area goes over buffer boundary,
			copy the data to the temp buffer.
		*/
		int len;
		len = runtime->buffer_size - chip->pos_out;
		urb_out->transfer_buffer = chip->wrap_out;

		if(len > 0) {
			memcpy(chip->wrap_out, runtime->dma_area + chip->pos_out * PODXTPRO_BYTES_PER_FRAME, len * PODXTPRO_BYTES_PER_FRAME);
			memcpy(chip->wrap_out + len * PODXTPRO_BYTES_PER_FRAME, runtime->dma_area, (urb_frames - len) * PODXTPRO_BYTES_PER_FRAME);
		}
		else
			dev_err(s2m(substream), "driver bug: len = %d\n", len);  /* this is somewhat paranoid */
	}
	else {
		/* set the buffer pointer */
		urb_out->transfer_buffer = runtime->dma_area + chip->pos_out * PODXTPRO_BYTES_PER_FRAME;
	}

	if((chip->pos_out += urb_frames) >= runtime->buffer_size)
		chip->pos_out -= runtime->buffer_size;

	urb_out->transfer_buffer_length = urb_size;
	urb_out->context = substream;
	change_volume(urb_out, chip->volume);

	if(usb_submit_urb(urb_out, GFP_ATOMIC) == 0)
		set_bit(index, &chip->active_urb_out);
	else
		dev_err(s2m(substream), "URB out #%d submission failed\n", index);
	
	spin_unlock_irqrestore(&chip->lock_audio_out, flags);
	return 0;
}

/*
	Submit all currently available playback URBs.
*/
static int submit_audio_out_all_urbs(snd_pcm_substream_t *substream)
{
	int ret, i;

	for(i = 0; i < PODXTPRO_ISO_BUFFERS; ++i)
		if((ret = submit_audio_out_urb(substream)) < 0)
			return ret;

	return 0;
}

/*
	Unlink all currently active playback URBs.
*/
static void unlink_audio_out_urbs(snd_podxtpro_audio_t *chip)
{
	unsigned int i;

	for(i = PODXTPRO_ISO_BUFFERS; i--;) {
		if(test_bit(i, &chip->active_urb_out)) {
			if(!test_and_set_bit(i, &chip->unlink_urb_out)) {
				struct urb *u = chip->urb_audio_out[i];
				u->transfer_flags |= URB_ASYNC_UNLINK;
				usb_unlink_urb(u);
			}
		}
	}
}

/*
	Wait until unlinking of all currently active playback URBs has been	finished.
*/
static void wait_clear_audio_out_urbs(snd_podxtpro_audio_t *chip)
{
	int timeout = HZ;
	unsigned int i;
	int alive;

	do {
		alive = 0;
		for (i = PODXTPRO_ISO_BUFFERS; i--;) {
			if (test_bit(i, &chip->active_urb_out))
				alive++;
		}
		if (! alive)
			break;
		set_current_state(TASK_UNINTERRUPTIBLE);
		schedule_timeout(1);
	} while (--timeout > 0);
	if (alive)
		snd_printk(KERN_ERR "timeout: still %d active urbs..\n", alive);

	chip->active_urb_out = 0;
	chip->unlink_urb_out = 0;
}

/*
	Unlink all currently active playback URBs, and wait for finishing.
*/
void unlink_wait_clear_audio_out_urbs(snd_podxtpro_audio_t *chip)
{
	unlink_audio_out_urbs(chip);
	wait_clear_audio_out_urbs(chip);
}

/*
	Callback for completed playback URB.
*/
static void audio_out_callback(struct urb *urb, struct pt_regs *regs)
{
	int i, index, length = 0, shutdown = 0;
	unsigned long flags;

	snd_pcm_substream_t *substream = (snd_pcm_substream_t *)urb->context;
	snd_podxtpro_audio_t *chip = snd_pcm_substream_chip(substream);
	snd_pcm_runtime_t *runtime = substream->runtime;

	/* find index of URB */
	for(index = PODXTPRO_ISO_BUFFERS; index--;)
		if(urb == chip->urb_audio_out[index])
			break;

	for(i = PODXTPRO_ISO_PACKETS; i--;)
		length += urb->iso_frame_desc[i].length;

	spin_lock_irqsave(&chip->lock_audio_out, flags);
	chip->pos_out_done += length / PODXTPRO_BYTES_PER_FRAME;

	if(chip->pos_out_done >= runtime->buffer_size)
		chip->pos_out_done -= runtime->buffer_size;
	
	clear_bit(index, &chip->active_urb_out);

	for(i = PODXTPRO_ISO_PACKETS; i--;)
		if(urb->iso_frame_desc[i].status == -ESHUTDOWN) {
			shutdown = 1;
			break;
		}

	if(test_bit(index, &chip->unlink_urb_out))  /* better use a chip->flag! */
		shutdown = 1;

	spin_unlock_irqrestore(&chip->lock_audio_out, flags);

	if(!shutdown) {
		submit_audio_out_urb(substream);

		if((chip->bytes_out += length) >= chip->period_out) {
			chip->bytes_out -= chip->period_out;
			snd_pcm_period_elapsed(substream);
		}
	}
}

/* open playback callback */
static int snd_podxtpro_playback_open(snd_pcm_substream_t *substream)
{
	snd_pcm_runtime_t *runtime = substream->runtime;
	runtime->hw = snd_podxtpro_playback_hw;
	return 0;
}

/* close playback callback */
static int snd_podxtpro_playback_close(snd_pcm_substream_t *substream)
{
	return 0;
}

/* hw_params playback callback */
static int snd_podxtpro_playback_hw_params(snd_pcm_substream_t *substream, snd_pcm_hw_params_t *hw_params)
{
	int ret;

	snd_podxtpro_audio_t *chip = snd_pcm_substream_chip(substream);

	if((ret = snd_pcm_lib_malloc_pages(substream, params_buffer_bytes(hw_params))) < 0)
		return ret;

	chip->period_out = params_period_bytes(hw_params);
	chip->wrap_out = kmalloc(PODXTPRO_ISO_PACKET_SIZE_NORMAL + PODXTPRO_ISO_PACKET_SIZE_LONG, GFP_KERNEL);

	if(!chip->wrap_out) {
		dev_err(s2m(substream), "cannot malloc wrap_out\n");
		return -ENOMEM;
	}

	return 0;
}

/* hw_free playback callback */
static int snd_podxtpro_playback_hw_free(snd_pcm_substream_t *substream)
{
	snd_podxtpro_audio_t *chip = snd_pcm_substream_chip(substream);
	unlink_wait_clear_audio_out_urbs(chip);

	if(chip->wrap_out) {
		kfree(chip->wrap_out);
		chip->wrap_out = 0;
	}

	return snd_pcm_lib_free_pages(substream);
}

/* prepare playback callback */
static int snd_podxtpro_playback_prepare(snd_pcm_substream_t *substream)
{
	snd_podxtpro_audio_t *chip = snd_pcm_substream_chip(substream);
	chip->pos_out = 0;
	chip->pos_out_done = 0;
	chip->bytes_out = 0;
	unlink_wait_clear_audio_out_urbs(chip);
	return 0;
}

/* trigger playback callback */
int snd_podxtpro_playback_trigger(snd_pcm_substream_t *substream, int cmd)
{
	snd_podxtpro_audio_t *chip = snd_pcm_substream_chip(substream);
	int err;

	switch (cmd) {
	case SNDRV_PCM_TRIGGER_START:
		if(!chip->running_playback) {
			err = submit_audio_out_all_urbs(substream);
			if(err < 0) return err;
			chip->running_playback = 1;
		}

		break;

	case SNDRV_PCM_TRIGGER_STOP:
		if(chip->running_playback) {
			unlink_audio_out_urbs(chip);
			chip->running_playback = 0;
		}

		break;

	default:
		return -EINVAL;
	}

	return 0;
}

/* playback pointer callback */
static snd_pcm_uframes_t
snd_podxtpro_playback_pointer(snd_pcm_substream_t *substream)
{
	snd_podxtpro_audio_t *chip = snd_pcm_substream_chip(substream);
	return chip->pos_out_done;
}

/* playback operators */
snd_pcm_ops_t snd_podxtpro_playback_ops = {
	.open =        snd_podxtpro_playback_open,
	.close =       snd_podxtpro_playback_close,
	.ioctl =       snd_pcm_lib_ioctl,
	.hw_params =   snd_podxtpro_playback_hw_params,
	.hw_free =     snd_podxtpro_playback_hw_free,
	.prepare =     snd_podxtpro_playback_prepare,
	.trigger =     snd_podxtpro_trigger,
	.pointer =     snd_podxtpro_playback_pointer,
};

int create_audio_out_urbs(snd_podxtpro_audio_t *chip)
{
	int i;
	usb_podxtpro_t *podxtpro = chip->podxtpro;

	/* create audio URBs and fill in constant values: */
	for(i = 0; i < PODXTPRO_ISO_BUFFERS; ++i)
		{
			struct urb *urb;

			/* URB for audio out: */
			urb = chip->urb_audio_out[i] = usb_alloc_urb(PODXTPRO_ISO_PACKETS, GFP_KERNEL);

			if(urb == NULL) {
				dev_err(podxtpro->msgdev, "Out of memory\n");
				return -ENOMEM;
			}

			urb->dev = podxtpro->usbdev;
			urb->pipe = usb_sndisocpipe(podxtpro->usbdev, PODXTPRO_EP_WRITE_AUDIO & USB_ENDPOINT_NUMBER_MASK);
			urb->transfer_flags = URB_ISO_ASAP;
			urb->start_frame = -1;
			urb->number_of_packets = PODXTPRO_ISO_PACKETS;
			urb->interval = PODXTPRO_ISO_INTERVAL;
			urb->error_count = 0;
			urb->complete = audio_out_callback;
		}

	return 0;
}
